class NoTagsFoundError(Exception):
    ...


class TooManyTagsError(Exception):
    ...


class NoKnownRegistryError(Exception):
    ...


class NoKnownCiServerError(Exception):
    ...


class NoKnownDnsServerError(Exception):
    ...
